#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <unistd.h>
#include <threads.h>
#include <stdatomic.h>

#include "states.h"
#include "open_gui.h"
#include "unix_sockets.h"
#include "melon_vpn_sock.h"
#include "melon_vpn_about.h"

#ifndef PROJ_VERSION
  #define PROJ_VERSION "unknown"
#endif

States currentstate;
char *currentpeers;
bool currentisrestart;
char *version = PROJ_VERSION;
bool is_gui = false;
thrd_t listenthr;
atomic_bool listenserverready = false;

int send_to_daemon(char*, ssize_t);
int refresh_vpn();
int status_vpn();
int peers_vpn();
int start_vpn();
int stop_vpn();
int restart_conf_vpn(bool);

int start_client_receiver(void*);
int listen_client_receiver(int);

void start_listen_thread();
void wait_listen_thread();

void update_state(States state) {
  currentstate = state;
}

bool incoming_state(char *state) {
  currentstate = enum_state(state);
  if(is_gui) {
    gui_trigger_state(currentstate);
    return false;
  } else {
    return true;
  }
}

bool incoming_peers(char *peers) {
  currentpeers = peers;
  if(is_gui) {
    gui_trigger_peers(peers);
    return false;
  }
  return true;
}

bool incoming_isrestart(char *isrestart) {
  currentisrestart = strcmp(isrestart,"on")==0;
  if(is_gui) {
    gui_trigger_isrestart(currentisrestart);
    return false;
  }
  return true;
}

int is_arg(int ac, char **argv, char *arg) {
  if (ac < 2) {
    return 0;
  }
  for(int x=1; x < ac; x++) {
    if (0 == strcmp(argv[x], arg)) {
      return x;  // return position of arg
    }
  }
  return 0;  // arg not present
}

void display_version(bool is_debug) {
  if(is_debug) printf("v%s-beta\n", version);
  else printf("v%s\n", version);
}

void display_about(bool is_debug) {
  printf("%s\n", about_name);
  display_version(is_debug);
  printf("%s\n", about_copyright);
  printf("%s\n", about_url);
  printf("%s\n", about_website);
  printf("Author: %s\n", about_author);
  printf("Artist: %s\n", about_artist);
}

void display_help(bool is_debug) {
  printf("Melon VPN Help\n");
  printf("-h, --help          Display this help message\n");
  printf("--about             About this program\n");
  printf("connect, -c         Connect to VPN\n");
  printf("disconnect, -d      Disconnect from VPN\n");
  printf("restart <on/off>,   Enable or disable restart mode\n");
  printf("     -r <on/off>\n");
}

int main (int argc, char **argv) {
  int status;
  currentstate = DISCONNECTED;

  bool is_debug = is_arg(argc, argv, "--debug");

  if(is_arg(argc, argv, "--help") || is_arg(argc, argv, "-h")) {
    display_help(is_debug);
    return 0;
  }

  if(is_arg(argc, argv, "--version") || is_arg(argc, argv, "-v")) {
    display_version(is_debug);
    return 0;
  }

  if(is_arg(argc, argv, "--about")) {
    display_about(is_debug);
    return 0;
  }

  start_listen_thread();

  struct timespec *t = malloc(sizeof t);
  t->tv_nsec = 1000;
  while(!listenserverready) {
    thrd_sleep(t,NULL);
  }

  if(is_arg(argc, argv, "--gui")) {
    is_gui = true;
    open_gui(argc, argv, is_debug, &start_vpn, &stop_vpn, &restart_conf_vpn, &refresh_vpn);
    is_gui = false;
    return 0;
  }

  if(argc==2) {
    if(strcmp(argv[1],"status")==0 || strcmp(argv[1],"-s")==0) {
      status_vpn();
    } else if(strcmp(argv[1],"peers")==0 || strcmp(argv[1],"-p")==0) {
      peers_vpn();
    } else if(strcmp(argv[1],"connect")==0 || strcmp(argv[1],"-c")==0) {
      start_vpn();
    } else if(strcmp(argv[1],"disconnect")==0 || strcmp(argv[1],"-d")==0) {
      stop_vpn();
    }
  } else if(argc==3) {
    if(strcmp(argv[1],"restart")==0 || strcmp(argv[1],"-r")==0) {
      if(strcmp(argv[2],"enable")==0 || strcmp(argv[2],"on")==0) {
        restart_conf_vpn(true);
      } else if(strcmp(argv[2],"disable")==0 || strcmp(argv[2],"off")==0) {
        restart_conf_vpn(false);
      }
    }
  }

  return status;
}

void start_listen_thread() {
  thrd_create(&listenthr, start_client_receiver, NULL);
}

void wait_listen_thread() {
  thrd_join(listenthr,NULL);
}

int send_to_daemon(char *buf, ssize_t n) {
  int sfd = unix_socket_client_start(SV_DAEMON_SOCK_PATH);
  if(sfd==-1) {
    fprintf(stderr, "Make sure the daemon is running before trying to use it.\n");
    return 1;
  }

  write(sfd, buf, n);
  close(sfd);

  fprintf(stderr, "Closed connection to socket\n");
  return 0;
}

int start_client_server() {
  int sfd = unix_socket_server_start(SV_CLIENT_SOCK_PATH);

  listenserverready = true;

  ssize_t numRead;
  int MAX_PEERS_LEN = 1000;
  char msg[MAX_PEERS_LEN];
  char buf[BUF_SIZE];
  int total = 0;

  for(;;) {
    int cfd = accept(sfd, NULL, NULL);

    strcpy(msg, "");
    strcpy(buf, "");
    total = 0;

    while ((numRead = read(cfd, buf, BUF_SIZE)) > 0) {
      if(total > MAX_PEERS_LEN-1) break;
      total+=numRead;
      strcat(msg, buf);
      msg[MAX_PEERS_LEN-1]=0;
    }

    fprintf(stderr, "Incoming request: %s\n",msg);

    char *a = strtok(msg,"\1");
    if(strcmp(a,"client")==0) {
      char *b = strtok(NULL,"\1");
      if(strcmp(b,"status")==0) {
        char *c = strtok(NULL,"\1");
        if(incoming_state(c)) break;
      } else if(strcmp(b,"peers")==0) {
        char *c = strtok(NULL,"\1");
        if(incoming_peers(c)) break;
      } else if(strcmp(b,"isrestart")==0) {
        char *c = strtok(NULL,"\1");
        if(incoming_isrestart(c)) break;
      }
    }
  }

  return 0;
}

int start_client_receiver(void *thr_data) {
  (void)thr_data;

  start_client_server();
  return 0;
}

int status_vpn() {
  char *msg = "daemon\1status";
  send_to_daemon(msg, strlen(msg));
  wait_listen_thread();
  printf("Melon VPN status: %s\n", string_state(currentstate));
  return 0;
}

int peers_vpn() {
  char *msg = "daemon\1peers";
  send_to_daemon(msg, strlen(msg));
  wait_listen_thread();
  printf("Melon VPN peers:\n%s\n", currentpeers);
  return 0;
}

int start_vpn() {
  char *msg = "daemon\1connect";
  return send_to_daemon(msg, strlen(msg));
}

int stop_vpn() {
  char *msg = "daemon\1disconnect";
  return send_to_daemon(msg, strlen(msg));
}

int refresh_vpn() {
  char *msg = "daemon\1status";
  int a=send_to_daemon(msg, strlen(msg));

  char *msg2 = "daemon\1peers";
  int b=send_to_daemon(msg2, strlen(msg2));

  char *msg3 = "daemon\1isrestart";
  int c=send_to_daemon(msg3, strlen(msg3));

  return a+b+c;
}

int restart_conf_vpn(bool enable) {
  char *msgprefix = "daemon\1restart\1";
  int prefixlen = strlen(msgprefix);
  int msglen = prefixlen + (enable ? 2 : 3);

  char msg[msglen+1];
  strcpy(msg, msgprefix);
  msg[prefixlen] = 0;

  strcat(msg, enable ? "on" : "off");
  msg[msglen] = 0;

  return send_to_daemon(msg, msglen+1);
}
